/***************************************************************************
 *   Copyright (C) 2004 by Work@                                           *
 *   jstubbs@work-at.co.jp                                                 *
 *                                                                         *
 *   This program may be distributed under the terms of the Q Public       *
 *   License as defined by Trolltech AS of Norway and appearing in the     *
 *   file LICENSE.QPL included in the packaging of this file.              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
 ***************************************************************************/

#include "cache.h"

#include <qdom.h>
#include <qfile.h>
#include <qt.h>

#define READ_TIMEOUT 10000

class Parser
{
	public:
		Parser(const QByteArray& data)
		{
			QString errormsg;
			int errrow;
			int errcol;
			if (!this->doc.setContent(data, false, &errormsg, &errrow, &errcol)) {
				valid = false;
				return;
			}
			
			valid = true;
			
			QDomNodeList channels = this->doc.documentElement().elementsByTagName("channel");
			this->channel = parse(channels.item(0).toElement());
			
			QDomNodeList items = this->doc.documentElement().elementsByTagName("item");
			for (unsigned int i = 0; i < items.count(); i++)
				this->headlines.append(parse(items.item(i).toElement()));
		}
		
		bool isValid() const
		{
			return valid;
		}
		
		ItemList items() const
		{
			ItemList items;
			
			for (unsigned int i = 0; i < this->headlines.count(); i++) {
				Item item;
				item.channel = channel;
				item.headline = headlines[i];
				items.append(item);
			}
			
			return items;
		}
		
		Link thechannel() const
		{
			return channel;
		}
		
	private:
		Link parse(QDomElement element) const
		{
			Link link;
			QDomNodeList children;
			
			children = element.elementsByTagName("link");
			if (children.count() != 0)
				link.url = children.item(0).toElement().text();
			
			children = element.elementsByTagName("title");
			if (children.count() != 0)
				link.title = children.item(0).toElement().text();
			
			children = element.elementsByTagName("description");
			if (children.count() != 0)
				link.comment = children.item(0).toElement().text();
			
			return link;
		}
		
		QDomDocument doc;
		bool valid;
		Link channel;
		LinkList headlines;
};

Cache::Cache(QObject* parent, const char* name)
	: QObject(parent, name)
{
	http = new QHttp(this);
	connect(http, SIGNAL(dataReadProgress(int, int)), this, SLOT(progress(int, int)));
	connect(http, SIGNAL(responseHeaderReceived(const QHttpResponseHeader&)),
	        this, SLOT(receiveResponse(const QHttpResponseHeader&)));
	connect(http, SIGNAL(readyRead(const QHttpResponseHeader&)),
	        this, SLOT(checkData(const QHttpResponseHeader&)));
	connect(http, SIGNAL(requestFinished(int, bool)), this, SLOT(complete(int, bool)));
	connect(http, SIGNAL(done(bool)), this, SLOT(fetchNext(bool)));
	connect(&timer, SIGNAL(timeout()), this, SLOT(http_timeout()));
	
	this->timeout = 0;
	this->currentid = -1;
	this->lasthost = "";
}

void
Cache::invalidate()
{
	QStringList keys = this->entries.keys();
	for (unsigned int i = 0; i < keys.count(); i++) {
		CacheEntry entry = this->entries[keys[i]];
		if (entry.lastmodified == "N/A") {
			this->entries.remove(keys[i]);
		} else {
			entry.time = 0;
			this->entries[keys[i]] = entry;
		}
	}
}

void
Cache::setTimeout(int timeout)
{
	this->timeout = timeout;
}

void
Cache::retrieve(QString url, bool check)
{
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::retrieve(url='"+url+"', check="+(check? "true":"false")+")");
	QString host;
	int port;
	QString path;
	QString newurl;
	parse(url, &host, &port, &path, &newurl);
	
	if (url != newurl) {
		qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit urlModified(url='"+url+"', newurl='"+newurl+"')");
		emit urlModified(url, newurl);
		url = newurl;
	}
	
	if (this->requests.contains(url)) {
		qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit progress(url='"+url+"', 0)");
		emit progress(url, 0);
		return;
	}
	
	if (this->entries.contains(url)) {
		if (entries[url].time + timeout * 60 > QDateTime::currentDateTime().toTime_t()) {
			qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit rssAvailable(url='"+url+"', items=...)");
			emit rssAvailable(url, this->entries[url].items);
			return;
		}
	}
	
	if (this->requests.count() == 0) {
		qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit cacheUpdating()");
		emit cacheUpdating();
	}
	
	this->requests.append(url);
	
	if (check && this->requests.count() == 1)
		fetchNext();
}

void
Cache::fetchNext(bool someval)
{
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::fetchNext(someval="+QString(someval? "true":"false")+")");
	if (requests.count() == 0) {
		qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit cacheCurrent()");
		emit cacheCurrent();
		return;
	}
	
	if (currentid != -1)
		return;
	
	QString url = requests[0];
	QString host;
	int port;
	QString path;
	QString newurl;
	parse(url, &host, &port, &path, &newurl);
	
	QHttpRequestHeader request;
	request.setValue("Host", host);
#ifdef Q_WS_MACX
	request.setValue("User-Agent", "RabbitTicker/1.0(MacOS)");
#else
	request.setValue("User-Agent", "RabbitTicker/1.0(Linux)");
#endif
	request.setRequest("GET", path);
	if (this->entries.contains(url)) {
		if (this->entries[url].lastmodified != "N/A") {
			request.setRequest("HEAD", path);
			request.setValue("If-Modified-Since", this->entries[url].lastmodified);
		}
	}
	
	response = QHttpResponseHeader();
	data.resize(0);
	
	if (lasthost != host) {
		currentid = this->http->setHost(host, port);
		qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  QHttp::setHost(host='"+host+"', port="+QString::number(port)+") ==> "+QString::number(currentid));
		lasthost = host;
		return;
	}
	
	currentid = this->http->request(request);
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  QHttp::request( QHttpRequestHeader(Host='"+request.value("Host")+
	       "', Request='"+request.method()+" "+request.path()+"') ) ==> "+QString::number(currentid));
	
	emit progress(url, 0);
	timer.stop();
	timer.start(READ_TIMEOUT, TRUE);
}

void
Cache::receiveResponse(const QHttpResponseHeader& header)
{
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::receiveResponse( QHttpResponseHeader(statusCode="+QString::number(header.statusCode())+") )");
	timer.stop();
	timer.start(READ_TIMEOUT, TRUE);
	if (http->currentId() == currentid)
		response = header;
}

void
Cache::checkData(const QHttpResponseHeader& header)
{
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::checkData( QHttpResponseHeader(statusCode="+QString::number(header.statusCode())+") )");
	timer.stop();
	timer.start(READ_TIMEOUT, TRUE);
	QByteArray newdata = http->readAll();
	if (data.count() == 0) {
		if (QCString(newdata).find("<rss", 0, false) == -1 && QCString(newdata).find("<rdf", 0, false) == -1) {
			qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  QHttp::abort()");
			this->http->abort();
			return;
		}
	}
	int index = data.count();
	data.resize(data.count() + newdata.count());
	for (unsigned int i = 0; i < newdata.count(); i++)
		data[index+i] = newdata[i];
}

void
Cache::http_timeout()
{
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::http_timeout()");
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  QHttp::abort()");
	this->http->abort();
}

void
Cache::complete(int id, bool failure)
{
	qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::complete(id="+QString::number(id)+", failure="+(failure? "true":"false")+")");
	
	if (id != currentid) {
		qDebug("id != currentid");
		return;
	}
	
	timer.stop();

	QString url = requests[0];
	
	currentid = -1;
	
	if (!failure && !response.isValid()) {
		qDebug("!failure && !response.isValid()");
		return;
	}
	
	qDebug("response.statusCode() == "+QString::number(response.statusCode()));
	qDebug("failure == "+QString(failure? "true":"false"));
	qDebug("retry == "+QString(retry? "true":"false"));
	
	if (response.statusCode() == 301 || response.statusCode() == 302 || !failure || retry) {
		this->requests.remove(url);
		retry = false;
	} else {
		retry = true;
	}
	
	bool wasmapped = false;
	if (mapping.contains(url)) {
		QString oldurl = mapping[url];
		if (!retry)
			mapping.remove(url);
		url = oldurl;
		wasmapped = true;
	}
	
	if (entries.contains(url)) {
		if (response.statusCode() == 304) {
			CacheEntry entry = this->entries[url];
			entry.time = QDateTime::currentDateTime().toTime_t();
			if (response.hasKey("last-modified") && !wasmapped)
				entry.lastmodified = response.value("last-modified");
			else
				entry.lastmodified = "N/A";
			this->entries[url] = entry;
			qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit rssAvailable(url='"+url+"', items=...)");
			emit rssAvailable(url, this->entries[url].items);
			emit rssTitle(url, entry.items[0].channel.title);
		} else {
			entries.remove(url);
			retrieve(url, false);
		}
	} else {
		if (response.statusCode() == 200) {
			Parser parser(data);
			ItemList items = parser.items();
			
			if (!parser.isValid()) {
				qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit rssInvalid(url='"+url+"')");
				emit rssInvalid(url);
			} else if (items.count() == 0) {
				qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit rssAvailable(url='"+url+"', items=...)");
				emit rssAvailable(url, items);
				emit rssTitle(url, parser.thechannel().title);
			} else {
				CacheEntry entry;
				entry.items = items;
				entry.time = QDateTime::currentDateTime().toTime_t();
				if (response.hasKey("last-modified") && !wasmapped)
					entry.lastmodified = response.value("last-modified");
				else
					entry.lastmodified = "N/A";
				this->entries[url] = entry;
				qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit rssAvailable(url='"+url+"', items=...)");
				emit rssAvailable(url, items);
				emit rssTitle(url, parser.thechannel().title);
			}
		} else if (response.statusCode() == 301 || response.statusCode() == 302) {
			QString alturl = response.value("location");
			if (!alturl.startsWith("http://")) {
				QString newurl;
				QString host;
				int port;
				QString path;
				parse(url, &host, &port, &path, &newurl);
				if (port == 80)
					alturl = "http://" + host + newurl;
				else
					alturl = "http://" + host + ":" + QString::number(port) + newurl;
			}
			if (response.statusCode() == 301) {
				qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit urlModified(url='"+url+"', newurl='"+alturl+"')");
				emit urlModified(url, alturl);
			} else {
				mapping[alturl] = url;
			}
			retrieve(alturl, false);
		} else if (!retry) {
			qDebug(QString::number(QDateTime::currentDateTime().toTime_t())+"  Cache::emit urlInvalid(url='"+url+"')");
			emit urlInvalid(url);
		}
	}
}

void
Cache::parse(QString url, QString* host, int* port, QString* path, QString* newurl)
{
	if (url.startsWith("http://"))
		url = url.mid(7);
	if (url.contains("/")) {
		*host = url.left(url.find("/"));
		*path = url.mid(host->length());
	} else {
		*host = url;
		*path = "/";
	}
	*port = 80;
	if (host->contains(":")) {
		bool ok;
		*port = host->mid(host->find(":")+1).toInt(&ok);
		if (!ok)
			*port = 80;
		*host = host->left(host->find(":"));
	}
	if (*port == 80)
		*newurl = "http://" + *host + *path;
	else
		*newurl = "http://" + *host + ":" + QString::number(*port) + *path;
}

void
Cache::progress(int done, int total)
{
	QString url = requests[0];
	
	if (total == 0)
		emit progress(url, 99);
	else
		emit progress(url, done * 100 / total);
}
