/***************************************************************************
 *   Copyright (C) 2004 by Work@                                           *
 *   jstubbs@work-at.co.jp                                                 *
 *                                                                         *
 *   This program may be distributed under the terms of the Q Public       *
 *   License as defined by Trolltech AS of Norway and appearing in the     *
 *   file LICENSE.QPL included in the packaging of this file.              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
 ***************************************************************************/

#include <qpixmap.h>
#include <qpainter.h>
#include <qtooltip.h>

#include "ticker.h"

const int TIMERFREQ = 10;

TickerDisplay::TickerDisplay(QWidget* parent, const char* name)
	: QWidget(parent, name)
{
	connect(&this->timer, SIGNAL(timeout()), this, SLOT(scroll()));
	this->setMouseTracking(true);
	this->timeout = 1000;
	this->scrolling = false;
	this->underlining = false;
	this->tooltips = true;
	this->left = true;
	this->mousein = false;
	this->mousepos = QPoint(-1, -1);
	this->bg = QColor(255, 255, 255);
	this->fg = QColor(0, 0, 0);
}

void
TickerDisplay::scrollLeft(Link link)
{
	if (this->tooltips)
		QToolTip::add(this, link.comment);
	
	if (!this->scrolling) {
		this->links.clear();
		this->positions.clear();
	} else if (!this->left) {
		while (this->positions.count() > 2) {
			this->links.pop_front();
			this->positions.pop_front();
		}
		this->left = true;
	}
	
	if (this->positions.count() == 0) {
		this->positions.push_back(0);
		this->links.push_back(link);
		this->repaint();
	} else {
		if (this->links.last() != link) {
			this->links.push_back(link);
			this->positions.push_back(this->positions.last() + this->width());
		}
		this->calcincrement((int)this->positions.first());
		this->repaint();
		this->timer.start(TIMERFREQ);
	}
}

void
TickerDisplay::scrollRight(Link link)
{
	if (this->tooltips)
		QToolTip::add(this, link.comment);
	
	if (!scrolling) {
		this->links.clear();
		this->positions.clear();
	} else if (this->left) {
		while (this->positions.count() > 2) {
			this->links.pop_back();
			this->positions.pop_back();
		}
		this->left = false;
	}
	
	if (this->positions.count() == 0) {
		this->positions.push_front(0);
		this->links.push_front(link);
		this->repaint();
	} else {
		if (this->links.first() != link) {
			this->links.push_front(link);
			this->positions.push_front(this->positions.first() - this->width());
		}
		this->calcincrement((int)this->positions.last());
		this->repaint();
		this->timer.start(TIMERFREQ);
	}
}

void
TickerDisplay::setFont(QFont font)
{
	this->font = font;
	this->repaint();
}

void
TickerDisplay::setBgColor(QColor bg)
{
	if (this->bg != bg) {
		this->bg = bg;
		this->repaint();
	}
}

void
TickerDisplay::setFgColor(QColor fg)
{
	if (this->fg != fg) {
		this->fg = fg;
		this->repaint();
	}
}

void
TickerDisplay::setTimeout(int timeout)
{
	this->timeout = timeout;
}

void
TickerDisplay::enableUnderlining(bool enabled)
{
	if (this->underlining != enabled) {
		this->underlining = enabled;
		if (this->mousein)
			this->repaint();
	}
}

void
TickerDisplay::enableScrolling(bool enabled)
{
	if (this->scrolling != enabled) {
		this->scrolling = enabled;
		if (this->scrolling)
			this->repaint();
	}
}

void
TickerDisplay::enableTooltips(bool enabled)
{
	if (this->tooltips != enabled) {
		this->tooltips = enabled;
		QToolTip::remove(this);
		if (this->tooltips)
			this->repaint();
	}
}

void
TickerDisplay::reset()
{
	this->links.clear();
	this->positions.clear();
	this->repaint();
}

void
TickerDisplay::scroll()
{
	this->repaint();
	
	if (this->links.count() == 0)
		return;
	
	if (this->links.count() == 1 && this->positions[0] == 0) {
		this->timer.stop();
		return;
	}
	
	for (unsigned int i = 0; i < this->positions.count(); i++)
		this->positions[i] += this->increment;
	
	if (left) {
		while (this->links.count() > 1 && this->positions.first() <= -this->width()) {
			this->links.pop_front();
			this->positions.pop_front();
		}
	} else {
		while (this->links.count() > 1 && this->positions.last() >= this->width()) {
			this->links.pop_back();
			this->positions.pop_back();
		}
	}
	
	if (this->positions.count() == 1) {
		if (this->left && this->positions.first() < 0) {
			this->positions[0] = 0;
		} else if (!this->left && this->positions.first() > 0) {
			this->positions[0] = 0;
		}
	}
}

void
TickerDisplay::calcincrement(int exclude)
{
	double count = links.count() - 1;
	if (count == 0)
		count = 1;
	double iterations = (double)timeout / (double)TIMERFREQ;
	double totalwidth = count * this->width() - exclude;
	this->increment = totalwidth / iterations;
	if (this->left)
		this->increment = -this->increment;
}

void
TickerDisplay::repaint(bool erase)
{
	QWidget::repaint(erase);
}

void
TickerDisplay::paintEvent(QPaintEvent*)
{
	QPixmap pixmap(size());
	pixmap.fill(this->bg);
	
	this->undermouse = Link();
	
	if (this->links.count() != 0) {
		QPainter p(&pixmap);
		p.setPen(this->fg);
		
		font.setUnderline(false);
		p.setFont(font);
		
		QRect rect = pixmap.rect();
		rect.setWidth(this->width() * 2);
		
		for (unsigned int i = 0; i < this->links.count(); i++) {
			if (this->positions[i] < -this->width() || this->positions[i] > this->width())
				continue;
			
			rect.moveLeft((int)this->positions[i]);
			QString text = this->links[i].title;
			QString post = "";
			
			QRect textrect = p.boundingRect(rect, Qt::AlignLeft | Qt::AlignVCenter, text);
			while (textrect.width() >= width()) {
				text = text.left(text.length() - 1);
				post = "...";
				textrect = p.boundingRect(rect, Qt::AlignLeft | Qt::AlignVCenter, text + post);
			}
			text += post;
			if (textrect.contains(this->mousepos) && this->mousein)
				this->undermouse = links[i];
			
			font.setUnderline(this->underlining && this->mousein && textrect.contains(this->mousepos) && this->links[i].url != "");
			p.setFont(font);
			p.drawText(rect, Qt::AlignLeft | Qt::AlignVCenter, text);
		}
		
		p.end();
	}
	
	QPainter pr(this);
	pr.drawPixmap(rect(), pixmap);
	pr.end();
}

void
TickerDisplay::mousePressEvent(QMouseEvent* e)
{
	if (e->button() == Qt::LeftButton && this->undermouse.url != "") {
		emit urlRequest(this->undermouse.url);
		e->accept();
	} else {
		e->ignore();
	}
}

void
TickerDisplay::mouseMoveEvent(QMouseEvent* e)
{
	this->mousepos = e->pos();
	e->ignore();
	this->repaint();
}

void
TickerDisplay::enterEvent(QEvent*)
{
	this->mousein = true;
	this->repaint();
}

void
TickerDisplay::leaveEvent(QEvent*)
{
	this->mousein = false;
	this->repaint();
}

Ticker::Ticker(QWidget* parent, const char* name)
	: QVBox(parent, name)
{
	this->channel = new TickerDisplay(this);
	connect(this->channel, SIGNAL(urlRequest(QString)),
	        this, SIGNAL(urlRequest(QString)));
	this->headline = new TickerDisplay(this);
	connect(this->headline, SIGNAL(urlRequest(QString)),
	        this, SIGNAL(urlRequest(QString)));
}

void
Ticker::scrollLeft(Item item)
{
	this->channel->scrollLeft(item.channel);
	this->headline->scrollLeft(item.headline);
}

void
Ticker::scrollRight(Item item)
{
	this->channel->scrollRight(item.channel);
	this->headline->scrollRight(item.headline);
}

void
Ticker::setFont(QFont font)
{
	this->channel->setFont(font);
	this->headline->setFont(font);
}

void
Ticker::setBgColor(QColor bg)
{
	this->channel->setBgColor(bg);
	this->headline->setBgColor(bg);
}

void
Ticker::setFgColor(QColor fg)
{
	this->channel->setFgColor(fg);
	this->headline->setFgColor(fg);
}

void
Ticker::setTimeout(int timeout)
{
	this->channel->setTimeout(timeout);
	this->headline->setTimeout(timeout);
}

void
Ticker::enableScrolling(bool enabled)
{
	this->channel->enableScrolling(enabled);
	this->headline->enableScrolling(enabled);
}

void
Ticker::enableTooltips(bool enabled)
{
	this->channel->enableTooltips(enabled);
	this->headline->enableTooltips(enabled);
}

void
Ticker::enableUnderlining(bool enabled)
{
	this->channel->enableUnderlining(enabled);
	this->headline->enableUnderlining(enabled);
}

void
Ticker::reset()
{
	this->channel->reset();
	this->headline->reset();
}
